/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.geometry;

import eu.svjatoslav.sixth.e3d.renderer.octree.IntegerPoint;

import static java.lang.Math.*;

/**
 * Used to represent point in a 3D space or vector.
 *
 * @see Point2D
 */

public class Point3D implements Cloneable {

    // coordinates
    public double x, y, z;

    public Point3D() {
    }

    public Point3D(final double x, final double y, final double z) {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    public Point3D(final float x, final float y, final float z) {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    public Point3D(final int x, final int y, final int z) {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    public Point3D(IntegerPoint point) {
        this.x = point.x;
        this.y = point.y;
        this.z = point.z;
    }


    /**
     * Creates new current point by cloning coordinates from parent point.
     */
    public Point3D(final Point3D parent) {
        x = parent.x;
        y = parent.y;
        z = parent.z;
    }

    /**
     * Add other point to current point. Value of other point will not be changed.
     *
     * @param otherPoint point to add.
     * @return current point.
     */
    public Point3D add(final Point3D otherPoint) {
        x += otherPoint.x;
        y += otherPoint.y;
        z += otherPoint.z;
        return this;
    }

    /**
     * Add coordinates of current point to other point. Value of current point will not be changed.
     *
     * @return current point.
     */
    public Point3D addTo(final Point3D... otherPoints) {
        for (final Point3D otherPoint : otherPoints) otherPoint.add(this);
        return this;
    }

    /**
     * Create new point by cloning position of current point.
     *
     * @return newly created clone.
     */
    public Point3D clone() {
        return new Point3D(this);
    }

    /**
     * Copy coordinates from other point to current point. Value of other point will not be changed.
     */
    public Point3D clone(final Point3D otherPoint) {
        x = otherPoint.x;
        y = otherPoint.y;
        z = otherPoint.z;
        return this;
    }

    /**
     * Set current point coordinates to the middle point between two other points.
     *
     * @param p1 first point.
     * @param p2 second point.
     * @return current point.
     */
    public Point3D computeMiddlePoint(final Point3D p1, final Point3D p2) {
        x = (p1.x + p2.x) / 2d;
        y = (p1.y + p2.y) / 2d;
        z = (p1.z + p2.z) / 2d;
        return this;
    }

    /**
     * @return true if current point coordinates are equal to zero.
     */
    public boolean isZero() {
        return (x == 0) && (y == 0) && (z == 0);
    }

    public double getAngleXZ(final Point3D anotherPoint) {
        return Math.atan2(x - anotherPoint.x, z - anotherPoint.z);
    }

    public double getAngleYZ(final Point3D anotherPoint) {
        return Math.atan2(y - anotherPoint.y, z - anotherPoint.z);
    }

    public double getAngleXY(final Point3D anotherPoint) {
        return Math.atan2(x - anotherPoint.x, y - anotherPoint.y);
    }

    /**
     * Compute distance to another point.
     *
     * @param anotherPoint point to compute distance to.
     * @return distance to another point.
     */
    public double getDistanceTo(final Point3D anotherPoint) {
        final double xDelta = x - anotherPoint.x;
        final double yDelta = y - anotherPoint.y;
        final double zDelta = z - anotherPoint.z;

        return sqrt(((xDelta * xDelta) + (yDelta * yDelta) + (zDelta * zDelta)));
    }

    /**
     * @return length of current vector.
     */
    public double getVectorLength() {
        return sqrt(((x * x) + (y * y) + (z * z)));
    }

    /**
     * Invert current point coordinates.
     *
     * @return current point.
     */
    public Point3D invert() {
        x = -x;
        y = -y;
        z = -z;
        return this;
    }

    /**
     * Rotate current point around center point by angleXZ and angleYZ.
     * <p>
     * See also: <a href="https://marctenbosch.com/quaternions/">Let's remove Quaternions from every 3D Engine</a>
     *
     * @param center  center point.
     * @param angleXZ angle around XZ axis.
     * @param angleYZ angle around YZ axis.
     */
    public Point3D rotate(final Point3D center, final double angleXZ,
                          final double angleYZ) {
        final double s1 = sin(angleXZ);
        final double c1 = cos(angleXZ);

        final double s2 = sin(angleYZ);
        final double c2 = cos(angleYZ);

        x -= center.x;
        y -= center.y;
        z -= center.z;

        final double y1 = (z * s2) + (y * c2);
        final double z1 = (z * c2) - (y * s2);

        final double x1 = (z1 * s1) + (x * c1);
        final double z2 = (z1 * c1) - (x * s1);

        x = x1 + center.x;
        y = y1 + center.y;
        z = z2 + center.z;

        return this;
    }

    public Point3D rotate(final double angleXZ, final double angleYZ) {
        return rotate(new Point3D(0, 0, 0), angleXZ, angleYZ);
    }

    /**
     * Round current point coordinates to integer values.
     */
    public void roundToInteger() {
        x = (int) x;
        y = (int) y;
        z = (int) z;
    }

    /**
     * Scale down current point by factor.
     * All coordinates will be divided by factor.
     *
     * @param factor factor to scale by.
     * @return current point.
     */
    public Point3D scaleDown(final double factor) {
        x /= factor;
        y /= factor;
        z /= factor;
        return this;
    }

    /**
     * Scale up current point by factor.
     * All coordinates will be multiplied by factor.
     *
     * @param factor factor to scale by.
     * @return current point.
     */
    public Point3D scaleUp(final double factor) {
        x *= factor;
        y *= factor;
        z *= factor;
        return this;
    }

    /**
     * Set current point coordinates to given values.
     *
     * @param x X coordinate.
     * @param y Y coordinate.
     * @param z Z coordinate.
     */
    public void setValues(final double x, final double y, final double z) {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    /**
     * Subtract other point from current point. Value of other point will not be changed.
     *
     * @return current point.
     */
    public Point3D subtract(final Point3D otherPoint) {
        x -= otherPoint.x;
        y -= otherPoint.y;
        z -= otherPoint.z;
        return this;
    }

    @Override
    public String toString() {
        return "x:" + x + " y:" + y + " z:" + z;
    }

    /**
     * Translate current point along X axis by given increment.
     *
     * @return current point.
     */
    public Point3D translateX(final double xIncrement) {
        x += xIncrement;
        return this;
    }

    /**
     * Translate current point along Y axis by given increment.
     *
     * @return current point.
     */
    public Point3D translateY(final double yIncrement) {
        y += yIncrement;
        return this;
    }

    /**
     * Translate current point along Z axis by given increment.
     *
     * @return current point.
     */
    public Point3D translateZ(final double zIncrement) {
        z += zIncrement;
        return this;
    }

    /**
     * Here we assume that Z coordinate is distance to the viewer.
     * If Z is positive, then point is in front of the viewer, and therefore it is visible.
     *
     * @return point visibility status.
     */
    public boolean isVisible() {
        return z > 0;
    }

    /**
     * Resets point coordinates to zero along all axes.
     *
     * @return current point.
     */
    public Point3D zero() {
        x = 0;
        y = 0;
        z = 0;
        return this;
    }

}
