/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.gui;

import eu.svjatoslav.sixth.e3d.gui.humaninput.HIDEventTracker;
import eu.svjatoslav.sixth.e3d.gui.humaninput.KeyboardFocusStack;
import eu.svjatoslav.sixth.e3d.renderer.raster.ShapeCollection;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.Set;
import java.util.Timer;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Java Swing GUI panel that contains canvas for 3D rendering.
 * Usually it is used as a part of {@link ViewFrame}.
 */
public class ViewPanel extends JPanel implements ComponentListener {
    private static final long serialVersionUID = 1683277888885045387L;
    private final HIDEventTracker HIDEventTracker = new HIDEventTracker(this);
    private final KeyboardFocusStack keyboardFocusStack;
    private final Avatar avatar = new Avatar();
    private final ShapeCollection rootShapeCollection = new ShapeCollection();
    private final Set<ViewRenderListener> viewRenderListeners = ConcurrentHashMap.newKeySet();
    public Color backgroundColor = Color.BLACK;

    /**
     * Stores milliseconds when last frame was updated. This is needed to calculate time delta between frames.
     * Time delta is used to calculate smooth animation.
     */
    private long lastUpdateMillis = 0;

    /**
     * Timer that is used to update canvas at target FPS rate.
     */
    private Timer canvasUpdateTimer;

    private ViewUpdateTimerTask canvasUpdateTimerTask;
    private RenderingContext renderingContext = null;

    /**
     * Currently target frames per second rate for this view. Target FPS can be changed at runtime.
     * 3D engine tries to be smart and only repaints screen when there are visible changes.
     */
    private int targetFPS = 60;

    /**
     * Set to true if it is known than next frame reeds to be painted. Flag is cleared
     * immediately after frame got updated.
     */
    private boolean viewRepaintNeeded = true;

    public ViewPanel() {
        viewRenderListeners.add(avatar);
        viewRenderListeners.add(HIDEventTracker);

        keyboardFocusStack = new KeyboardFocusStack(this);

        initializePanelLayout();

        setFrameRate(targetFPS);

        addComponentListener(this);
    }

    public Avatar getAvatar() {
        return avatar;
    }

    public KeyboardFocusStack getKeyboardFocusStack() {
        return keyboardFocusStack;
    }

    public ShapeCollection getRootShapeCollection() {
        return rootShapeCollection;
    }

    public HIDEventTracker getHIDInputTracker() {
        return HIDEventTracker;
    }

    public void addViewUpdateListener(final ViewRenderListener listener) {
        viewRenderListeners.add(listener);
    }

    @Override
    public void componentHidden(final ComponentEvent e) {

    }

    @Override
    public void componentMoved(final ComponentEvent e) {

    }

    @Override
    public void componentResized(final ComponentEvent e) {
        viewRepaintNeeded = true;
    }

    @Override
    public void componentShown(final ComponentEvent e) {
        viewRepaintNeeded = true;
    }

    @Override
    public Dimension getMaximumSize() {
        return getPreferredSize();
    }

    @Override
    public Dimension getMinimumSize() {
        return getPreferredSize();
    }

    @Override
    public java.awt.Dimension getPreferredSize() {
        return new java.awt.Dimension(640, 480);
    }

    public RenderingContext getRenderingContext() {
        return renderingContext;
    }

    private void initializePanelLayout() {
        setFocusCycleRoot(true);
        setOpaque(true);
        setFocusable(true);
        setDoubleBuffered(false);
        setVisible(true);
        requestFocusInWindow();
    }

    private void renderFrame() {
        // paint root geometry collection to the offscreen render buffer
        clearCanvas();
        rootShapeCollection.paint(this, renderingContext);

        // draw rendered offscreen buffer to visible screen
        final Graphics graphics = getGraphics();
        if (graphics != null)
            graphics.drawImage(renderingContext.bufferedImage, 0, 0, null);
    }

    private void clearCanvas() {
        renderingContext.graphics.setColor(backgroundColor);
        renderingContext.graphics.fillRect(0, 0, getWidth(), getHeight());
    }

    /**
     * Calling these methods tells 3D engine that current 3D view needs to be
     * repainted on first opportunity.
     */
    public void repaintDuringNextViewUpdate() {
        viewRepaintNeeded = true;
    }

    /**
     * Set target frames per second rate for this view. Target FPS can be changed at runtime.
     * @param frameRate target frames per second rate for this view.
     */
    public void setFrameRate(final int frameRate) {
        if (canvasUpdateTimerTask != null) {
            canvasUpdateTimerTask.cancel();
            canvasUpdateTimerTask = null;

            canvasUpdateTimer.cancel();
            canvasUpdateTimer = null;
        }

        targetFPS = frameRate;

        if (frameRate <= 0) return;

        canvasUpdateTimer = new Timer();
        canvasUpdateTimerTask = new ViewUpdateTimerTask(this);

        // schedule timer task to run in frequency according to defined frame rate
        canvasUpdateTimer.schedule(canvasUpdateTimerTask, 0,
                1000 / frameRate);
    }

    /**
     * Stops rendering of this view.
     */
    public void stop() {
        setFrameRate(0);
    }

    /**
     * This method is executed by periodic timer task, in frequency according to
     * defined frame rate.
     * <p>
     * It tells view to update itself. View can decide if actual re-rendering of
     * graphics is needed.
     */
    void ensureThatViewIsUpToDate() {
        maintainRenderingContext();

        final int millisecondsPassedSinceLastUpdate = getMillisecondsPassedSinceLastUpdate();

        boolean renderFrame = notifyViewRenderListeners(millisecondsPassedSinceLastUpdate);

        if (viewRepaintNeeded) {
            viewRepaintNeeded = false;
            renderFrame = true;
        }

        // abort rendering if window size is invalid
        if ((getWidth() > 0) && (getHeight() > 0) && renderFrame) {
            renderFrame();
            viewRepaintNeeded = renderingContext.handlePossibleComponentMouseEvent();
        }

    }

    private void maintainRenderingContext() {
        int panelWidth = getWidth();
        int panelHeight = getHeight();

        if (panelWidth <= 0 || panelHeight <= 0) {
            renderingContext = null;
            return;
        }

        // create new rendering context if window size has changed
        if ((renderingContext == null)
                || (renderingContext.width != panelWidth)
                || (renderingContext.height != panelHeight)) {
            renderingContext = new RenderingContext(panelWidth, panelHeight);
        }

        renderingContext.prepareForNewFrameRendering();
    }

    private boolean notifyViewRenderListeners(int millisecondsPassedSinceLastUpdate) {
        boolean reRenderFrame = false;
        for (final ViewRenderListener listener : viewRenderListeners)
            if (listener.beforeRender(this, millisecondsPassedSinceLastUpdate))
                reRenderFrame = true;
        return reRenderFrame;
    }

    private int getMillisecondsPassedSinceLastUpdate() {
        final long currentTime = System.currentTimeMillis();

        if (lastUpdateMillis == 0)
            lastUpdateMillis = currentTime;

        final int millisecondsPassedSinceLastUpdate = (int) (currentTime - lastUpdateMillis);
        lastUpdateMillis = currentTime;
        return millisecondsPassedSinceLastUpdate;
    }

    public void addViewRenderListener(ViewRenderListener viewRenderListener) {
        viewRenderListeners.add(viewRenderListener);
    }

    public void removeViewRenderListener(ViewRenderListener viewRenderListener) {
        viewRenderListeners.remove(viewRenderListener);
    }

}
