/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.gui.textEditorComponent;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.gui.GuiComponent;
import eu.svjatoslav.sixth.e3d.gui.TextPointer;
import eu.svjatoslav.sixth.e3d.gui.ViewPanel;
import eu.svjatoslav.sixth.e3d.gui.humaninput.KeyboardHelper;
import eu.svjatoslav.sixth.e3d.math.Transform;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.textcanvas.TextCanvas;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

public class TextEditComponent extends GuiComponent implements ClipboardOwner {

    private static final long serialVersionUID = -7118833957783600630L;

    /**
     * Text rows that need to be repainted.
     */
    private final Set<Integer> dirtyRows = new HashSet<>();


    private final TextCanvas textCanvas;
    public int scrolledCharacters = 0, scrolledLines = 0;
    public boolean selecting = false;

    /**
     * Selection start and end pointers.
     */
    public TextPointer selectionStart = new TextPointer(0, 0);
    public TextPointer selectionEnd = new TextPointer(0, 0);


    public TextPointer cursorLocation = new TextPointer(0, 0);
    Page page = new Page();
    LookAndFeel lookAndFeel;

    /**
     * If true, the page will be repainted on the next update.
     */
    boolean repaintPage = false;

    public TextEditComponent(final Transform transform,
                             final ViewPanel viewPanel,
                             final Point2D sizeInWorldCoordinates,
                             LookAndFeel lookAndFeel) {
        super(transform, viewPanel, sizeInWorldCoordinates.to3D());

        this.lookAndFeel = lookAndFeel;
        final int columns = (int) (sizeInWorldCoordinates.x / TextCanvas.FONT_CHAR_WIDTH);
        final int rows = (int) (sizeInWorldCoordinates.y / TextCanvas.FONT_CHAR_HEIGHT);

        textCanvas = new TextCanvas(
                new Transform(),
                new TextPointer(rows, columns),
                lookAndFeel.foreground, lookAndFeel.background);

        textCanvas.setMouseInteractionController(this);

        repaintPage();
        addShape(textCanvas);
    }

    private void checkCursorBoundaries() {
        if (cursorLocation.column < 0)
            cursorLocation.column = 0;
        if (cursorLocation.row < 0)
            cursorLocation.row = 0;

        // ensure chat cursor stays within vertical editor boundaries by
        // vertical scrolling
        if ((cursorLocation.row - scrolledLines) < 0)
            scroll(0, cursorLocation.row - scrolledLines);

        if ((((cursorLocation.row - scrolledLines) + 1)) > textCanvas.getSize().row)
            scroll(0,
                    ((((((cursorLocation.row - scrolledLines) + 1) - textCanvas
                            .getSize().row)))));

        // ensure chat cursor stays within horizontal editor boundaries by
        // horizontal scrolling
        if ((cursorLocation.column - scrolledCharacters) < 0)
            scroll(cursorLocation.column - scrolledCharacters, 0);

        if ((((cursorLocation.column - scrolledCharacters) + 1)) > textCanvas
                .getSize().column)
            scroll((((((cursorLocation.column - scrolledCharacters) + 1) - textCanvas
                    .getSize().column))), 0);
    }

    /**
     * Clear text selection.
     */
    public void clearSelection() {
        selectionEnd = new TextPointer(selectionStart);
        repaintPage = true;
    }

    /**
     * Copies selected text to the clipboard.
     */
    public void copyToClipboard() {
        if (selectionStart.compareTo(selectionEnd) == 0)
            return;
        // System.out.println("Copy action.");
        final StringBuilder msg = new StringBuilder();

        ensureSelectionOrder();

        for (int row = selectionStart.row; row <= selectionEnd.row; row++) {
            final TextLine textLine = page.getLine(row);

            if (row == selectionStart.row) {
                if (row == selectionEnd.row)
                    msg.append(textLine.getSubString(selectionStart.column,
                            selectionEnd.column + 1));
                else
                    msg.append(textLine.getSubString(selectionStart.column,
                            textLine.getLength()));
            } else {
                msg.append('\n');
                if (row == selectionEnd.row)
                    msg.append(textLine
                            .getSubString(0, selectionEnd.column + 1));
                else
                    msg.append(textLine.toString());
            }
        }

        setClipboardContents(msg.toString());
    }

    public void cutToClipboard() {
        copyToClipboard();
        deleteSelection();
        repaintPage();
    }

    public void deleteSelection() {
        ensureSelectionOrder();
        int ym = 0;

        for (int line = selectionStart.row; line <= selectionEnd.row; line++) {
            final TextLine currentLine = page.getLine(line - ym);

            if (line == selectionStart.row) {
                if (line == selectionEnd.row)

                    currentLine.cutSubString(selectionStart.column,
                            selectionEnd.column);
                else if (selectionStart.column == 0) {
                    page.removeLine(line - ym);
                    ym++;
                } else
                    currentLine.cutSubString(selectionStart.column,
                            currentLine.getLength() + 1);
            } else if (line == selectionEnd.row)
                currentLine.cutSubString(0, selectionEnd.column);
            else {
                page.removeLine(line - ym);
                ym++;
            }
        }

        clearSelection();
        cursorLocation = new TextPointer(selectionStart);
    }

    /**
     * Ensures that {@link #selectionStart} is smaller than
     * {@link #selectionEnd}.
     */
    public void ensureSelectionOrder() {
        if (selectionStart.compareTo(selectionEnd) > 0) {
            final TextPointer temp = selectionEnd;
            selectionEnd = selectionStart;
            selectionStart = temp;
        }
    }

    public String getClipboardContents() {
        String result = "";
        final Clipboard clipboard = Toolkit.getDefaultToolkit()
                .getSystemClipboard();
        // odd: the Object param of getContents is not currently used
        final Transferable contents = clipboard.getContents(null);
        final boolean hasTransferableText = (contents != null)
                && contents.isDataFlavorSupported(DataFlavor.stringFlavor);
        if (hasTransferableText)
            try {
                result = (String) contents
                        .getTransferData(DataFlavor.stringFlavor);
            } catch (final UnsupportedFlavorException | IOException ex) {
                // highly unlikely since we are using a standard DataFlavor
                System.out.println(ex);
            }
        // System.out.println(result);
        return result;
    }

    /**
     * Place string into system clipboard so that it can be pasted into other
     * applications.
     */
    public void setClipboardContents(final String contents) {
        final StringSelection stringSelection = new StringSelection(contents);
        final Clipboard clipboard = Toolkit.getDefaultToolkit()
                .getSystemClipboard();
        clipboard.setContents(stringSelection, stringSelection);
    }

    public void goToLine(final int Line) {
        // markNavigationLocation(Line);
        scrolledLines = Line + 1;
        cursorLocation.row = Line + 1;
        cursorLocation.column = 0;
        repaintPage();
    }

    public void insertText(final String txt) {
        if (txt == null)
            return;

        for (final char c : txt.toCharArray()) {

            if (c == KeyboardHelper.DEL) {
                processDel();
                continue;
            }

            if (c == KeyboardHelper.ENTER) {
                processEnter();
                continue;
            }

            if (c == KeyboardHelper.BACKSPACE) {
                processBackspace();
                continue;
            }

            // type character
            if (KeyboardHelper.isText(c)) {
                page.insertCharacter(cursorLocation.row, cursorLocation.column,
                        c);
                cursorLocation.column++;
            }
        }
    }

    /**
     * Parse key presses.
     */
    @Override
    public boolean keyPressed(final KeyEvent event, final ViewPanel viewPanel) {
        super.keyPressed(event, viewPanel);

        processKeyEvent(event);

        markRowDirty();

        checkCursorBoundaries();

        repaintWhatNeeded();
        return true;
    }

    /**
     * Empty implementation of the ClipboardOwner interface.
     */
    @Override
    public void lostOwnership(final Clipboard aClipboard,
                              final Transferable aContents) {
        // do nothing
    }

    public void markRowDirty() {
        dirtyRows.add(cursorLocation.row);
    }

    public void pasteFromClipboard() {
        insertText(getClipboardContents());
    }

    private void processBackspace() {
        if (selectionStart.compareTo(selectionEnd) == 0) {
            // erase single character
            if (cursorLocation.column > 0) {
                cursorLocation.column--;
                page.removeCharacter(cursorLocation.row, cursorLocation.column);
                // System.out.println(lines.get(currentCursor.line).toString());
            } else if (cursorLocation.row > 0) {
                cursorLocation.row--;
                final int currentLineLength = page
                        .getLineLength(cursorLocation.row);
                cursorLocation.column = currentLineLength;
                page.getLine(cursorLocation.row)
                        .insertTextLine(currentLineLength,
                                page.getLine(cursorLocation.row + 1));
                page.removeLine(cursorLocation.row + 1);
                repaintPage = true;
            }
        } else {
            // dedent multiple lines
            ensureSelectionOrder();
            // scan if enough space exists
            for (int y = selectionStart.row; y < selectionEnd.row; y++)
                if (page.getLine(y).getIdent() < 4)
                    return;

            for (int y = selectionStart.row; y < selectionEnd.row; y++)
                page.getLine(y).cutFromBeginning(4);

            repaintPage = true;
        }
    }

    private void processCtrlCombinations(final int keyCode) {

        if ((char) keyCode == 'A') { // CTRL + A -- select all
            final int lastLineIndex = page.getLinesCount() - 1;
            selectionStart = new TextPointer(0, 0);
            selectionEnd = new TextPointer(lastLineIndex,
                    page.getLineLength(lastLineIndex));
            repaintPage();
        }

        // CTRL + X -- cut
        if ((char) keyCode == 'X')
            cutToClipboard();

        // CTRL + C -- copy
        if ((char) keyCode == 'C')
            copyToClipboard();

        // CTRL + V -- paste
        if ((char) keyCode == 'V')
            pasteFromClipboard();

        if (keyCode == 39) { // RIGHT
            // skip to the beginning of the next word

            for (int x = cursorLocation.column; x < (page
                    .getLineLength(cursorLocation.row) - 1); x++)
                if ((page.getChar(cursorLocation.row, x) == ' ')
                        && (page.getChar(cursorLocation.row, x + 1) != ' ')) {
                    // beginning of the next word is found
                    cursorLocation.column = x + 1;
                    return;
                }

            cursorLocation.column = page.getLineLength(cursorLocation.row);
            return;
        }

        if (keyCode == 37) { // Left

            // skip to the beginning of the previous word
            for (int x = cursorLocation.column - 2; x >= 0; x--)
                if ((page.getChar(cursorLocation.row, x) == ' ')
                        & (page.getChar(cursorLocation.row, x + 1) != ' ')) {
                    cursorLocation.column = x + 1;
                    return;
                }

            cursorLocation.column = 0;
        }
    }

    public void processDel() {
        if (selectionStart.compareTo(selectionEnd) == 0) {
            // is there still some text right to the cursor ?
            if (cursorLocation.column < page.getLineLength(cursorLocation.row))
                page.removeCharacter(cursorLocation.row, cursorLocation.column);
            else {
                page.getLine(cursorLocation.row).insertTextLine(
                        cursorLocation.column,
                        page.getLine(cursorLocation.row + 1));
                page.removeLine(cursorLocation.row + 1);
                repaintPage = true;
            }
        } else {
            deleteSelection();
            repaintPage = true;
        }
    }

    private void processEnter() {
        final TextLine currentLine = page.getLine(cursorLocation.row);
        // move everything right to the cursor into new line
        final TextLine newLine = currentLine.getSubLine(cursorLocation.column,
                currentLine.getLength());
        page.insertLine(cursorLocation.row + 1, newLine);

        // trim existing line
        page.getLine(cursorLocation.row).cutUntilEnd(cursorLocation.column);
        repaintPage = true;

        cursorLocation.row++;
        cursorLocation.column = 0;
    }

    private void processKeyEvent(final KeyEvent event) {
        final int modifiers = event.getModifiersEx();
        final int keyCode = event.getKeyCode();
        final char keyChar = event.getKeyChar();

        // System.out.println("Keycode:" + keyCode s+ ", keychar:" + keyChar);

        if (KeyboardHelper.isAltPressed(modifiers))
            return;

        if (KeyboardHelper.isCtrlPressed(modifiers)) {
            processCtrlCombinations(keyCode);
            return;
        }

        if (keyCode == KeyboardHelper.TAB) {
            processTab(modifiers);
            return;
        }

        clearSelection();

        if (KeyboardHelper.isText(keyCode)) {
            insertText(String.valueOf(keyChar));
            return;
        }

        if (KeyboardHelper.isShiftPressed(modifiers)) {
            if (!selecting)
                attemptSelectionStart:{

                    if (keyChar == 65535)
                        if (keyCode == 16)
                            break attemptSelectionStart;
                    if (((keyChar >= 32) & (keyChar <= 128)) | (keyChar == 10)
                            | (keyChar == 8) | (keyChar == 9))
                        break attemptSelectionStart;

                    selectionStart = new TextPointer(cursorLocation);
                    selectionEnd = selectionStart;
                    selecting = true;
                    repaintPage();
                }
        } else
            selecting = false;

        if (keyCode == KeyboardHelper.HOME) {
            cursorLocation.column = 0;
            return;
        }
        if (keyCode == KeyboardHelper.END) {
            cursorLocation.column = page.getLineLength(cursorLocation.row);
            return;
        }

        // process cursor keys
        if (keyCode == KeyboardHelper.DOWN) {
            markRowDirty();
            cursorLocation.row++;
            return;
        }

        if (keyCode == KeyboardHelper.UP) {
            markRowDirty();
            cursorLocation.row--;
            return;
        }

        if (keyCode == KeyboardHelper.RIGHT) {
            cursorLocation.column++;
            return;
        }

        if (keyCode == KeyboardHelper.LEFT) {
            cursorLocation.column--;
            return;
        }

        if (keyCode == KeyboardHelper.PGDOWN) {
            cursorLocation.row += textCanvas.getSize().row;
            repaintPage();
            return;
        }

        if (keyCode == KeyboardHelper.PGUP) {
            cursorLocation.row -= textCanvas.getSize().row;
            repaintPage = true;
        }

    }

    private void processTab(final int modifiers) {
        if (KeyboardHelper.isShiftPressed(modifiers)) {
            if (selectionStart.compareTo(selectionEnd) != 0) {
                // dedent multiple lines
                ensureSelectionOrder();

                identSelection:
                {
                    // check that indentation is possible
                    for (int y = selectionStart.row; y < selectionEnd.row; y++) {
                        final TextLine textLine = page.getLine(y);

                        if (!textLine.isEmpty())
                            if (textLine.getIdent() < 4)
                                break identSelection;
                    }

                    for (int y = selectionStart.row; y < selectionEnd.row; y++)
                        page.getLine(y).cutFromBeginning(4);
                }
            } else {
                // dedent current line
                final TextLine textLine = page.getLine(cursorLocation.row);

                if (cursorLocation.column >= 4)
                    if (textLine.isEmpty())
                        cursorLocation.column -= 4;
                    else if (textLine.getIdent() >= 4) {
                        cursorLocation.column -= 4;
                        textLine.cutFromBeginning(4);
                    }

            }

            repaintPage();

        } else if (selectionStart.compareTo(selectionEnd) != 0) {
            // ident multiple lines
            ensureSelectionOrder();
            for (int y = selectionStart.row; y < selectionEnd.row; y++)
                page.getLine(y).addIdent(4);

            repaintPage();
        }
    }

    public void repaintPage() {

        final int columnCount = textCanvas.getSize().column + 2;
        final int rowCount = textCanvas.getSize().row + 2;

        for (int row = 0; row < rowCount; row++)
            for (int column = 0; column < columnCount; column++) {
                final boolean isTabMargin = ((column + scrolledCharacters) % 4) == 0;

                if ((column == (cursorLocation.column - scrolledCharacters))
                        & (row == (cursorLocation.row - scrolledLines))) {
                    // cursor
                    textCanvas.setBackgroundColor(lookAndFeel.cursorBackground);
                    textCanvas.setForegroundColor(lookAndFeel.cursorForeground);
                } else if (new TextPointer(row + scrolledLines, column).isBetween(
                        selectionStart, selectionEnd)) {
                    // selected text
                    textCanvas.setBackgroundColor(lookAndFeel.selectionBackground);
                    textCanvas.setForegroundColor(lookAndFeel.selectionForeground);
                } else {
                    // normal text
                    textCanvas.setBackgroundColor(lookAndFeel.background);
                    textCanvas.setForegroundColor(lookAndFeel.foreground);

                    if (isTabMargin)
                        textCanvas
                                .setBackgroundColor(lookAndFeel.tabStopBackground);

                }

                final char charUnderCursor = page.getChar(row + scrolledLines,
                        column + scrolledCharacters);

                textCanvas.putChar(row, column, charUnderCursor);
            }

    }

    public void repaintRow(final int rowNumber) {
        // TODO: Optimize this. No need to repaint entire page.
        repaintPage();
    }

    private void repaintWhatNeeded() {
        if (repaintPage) {
            dirtyRows.clear();
            repaintPage();
            return;
        }

        dirtyRows.forEach(this::repaintRow);
        dirtyRows.clear();
    }

    /**
     * Scroll full page to given amount of lines or charancters.
     */
    public void scroll(final int charactersToScroll, final int linesToScroll) {
        scrolledLines += linesToScroll;
        scrolledCharacters += charactersToScroll;

        if (scrolledLines < 0)
            scrolledLines = 0;

        if (scrolledCharacters < 0)
            scrolledCharacters = 0;

        repaintPage = true;
    }

    public void setText(final String text) {
        // System.out.println("Set text:" + text);
        cursorLocation = new TextPointer(0, 0);
        scrolledCharacters = 0;
        scrolledLines = 0;
        selectionStart = new TextPointer(0, 0);
        selectionEnd = new TextPointer(0, 0);
        page = new Page();
        insertText(text);
        repaintPage();
    }

}
