/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.math;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;

/**
 * Vertex is a point where two or more lines, line segments, or rays come together.
 * In other words, it's a corner of a polygon, polyhedron, or other geometric shape.
 * For example, a triangle has three vertices, a square has four, and a cube has eight.
 */
public class Vertex {

    /**
     * Vertex coordinate in 3D space.
     */
    public Point3D coordinate;

    /**
     * Vertex coordinate relative to the viewer after transformation.
     * Visible vertices have positive z coordinate.
     * Viewer is located at (0, 0, 0).
     * No perspective correction is applied.
     */
    public Point3D transformedCoordinate;

    /**
     * Vertex coordinate in pixels relative to the top left corner of the screen after transformation
     * and perspective correction.
     */
    public Point2D onScreenCoordinate;


    /**
     * Coordinate within texture.
     */
    public Point2D textureCoordinate;


    /**
     * The frame number when this vertex was last transformed.
     */
    private int lastTransformedFrame;

    public Vertex() {
        this(new Point3D());
    }

    public Vertex(final Point3D location) {
        this(location, null);
    }

    public Vertex(final Point3D location, Point2D textureCoordinate) {
        coordinate = location;
        transformedCoordinate = new Point3D();
        onScreenCoordinate = new Point2D();
        this.textureCoordinate = textureCoordinate;
    }


    /**
     * Transforms vertex coordinate to calculate its location relative to the viewer.
     * It also calculates its location on the screen.
     *
     * @param transforms    Transforms pipeline.
     * @param renderContext Rendering context.
     */
    public void calculateLocationRelativeToViewer(final TransformsStack transforms,
                                                  final RenderingContext renderContext) {

        if (lastTransformedFrame == renderContext.frameNumber)
            return;

        lastTransformedFrame = renderContext.frameNumber;

        transforms.transform(coordinate, transformedCoordinate);

        onScreenCoordinate.x = ((transformedCoordinate.x / transformedCoordinate.z) * renderContext.zoom);
        onScreenCoordinate.y = ((transformedCoordinate.y / transformedCoordinate.z) * renderContext.zoom);
        onScreenCoordinate.add(renderContext.centerCoordinate);
    }
}
