/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.math.Vertex;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.TextureBitmap;

/**
 * Texture object that is always oriented towards the viewer.
 */
public class ForwardOrientedTexture extends AbstractCoordinateShape {

    private static final double SCALE_MULTIPLIER = 0.005;
    public final Texture texture;

    /**
     * Scale of the texture object.
     * <p>
     * Object rendered visible size on the screen depends on underlying texture size and scale.
     * <p>
     * 0 means that object will be infinitely small.
     * 1 in recommended value to maintain sharpness of the texture as seen by the viewer.
     */
    private double scale;

    public ForwardOrientedTexture(final Point3D point, final double scale,
                                  final Texture texture) {
        super(new Vertex(point));
        this.texture = texture;
        setScale(scale);
    }

    /**
     * Paint the texture on the screen (targetRenderingArea)
     *
     * @param targetRenderingArea the screen to paint on
     */
    @Override
    public void paint(final RenderingContext targetRenderingArea) {

        // distance from camera/viewer to center of the texture
        final double z = coordinates[0].transformedCoordinate.z;

        // compute forward oriented texture visible distance from center
        final double visibleHorizontalDistanceFromCenter = (targetRenderingArea.width
                * scale * texture.primaryBitmap.width) / z;

        final double visibleVerticalDistanceFromCenter = (targetRenderingArea.width
                * scale * texture.primaryBitmap.height) / z;

        // compute visible pixel density, and get appropriate bitmap
        final double zoom = (visibleHorizontalDistanceFromCenter * 2)
                / texture.primaryBitmap.width;

        final TextureBitmap textureBitmap = texture.getZoomedBitmap(zoom);

        final Point2D onScreenCoordinate = coordinates[0].onScreenCoordinate;

        // compute Y
        final int onScreenUncappedYStart = (int) (onScreenCoordinate.y - visibleVerticalDistanceFromCenter);
        final int onScreenUncappedYEnd = (int) (onScreenCoordinate.y + visibleVerticalDistanceFromCenter);
        final int onScreenUncappedHeight = onScreenUncappedYEnd - onScreenUncappedYStart;

        int onScreenCappedYStart = onScreenUncappedYStart;
        int onScreenCappedYEnd = onScreenUncappedYEnd;

        // cap Y to upper screen border
        if (onScreenCappedYStart < 0)
            onScreenCappedYStart = 0;

        // cap Y to lower screen border
        if (onScreenCappedYEnd > targetRenderingArea.height)
            onScreenCappedYEnd = targetRenderingArea.height;

        // compute X
        final int onScreenUncappedXStart = (int) (onScreenCoordinate.x - visibleHorizontalDistanceFromCenter);
        final int onScreenUncappedXEnd = (int) (onScreenCoordinate.x + visibleHorizontalDistanceFromCenter);
        final int onScreenUncappedWidth = onScreenUncappedXEnd - onScreenUncappedXStart;

        // cap X to left screen border
        int onScreenCappedXStart = onScreenUncappedXStart;
        if (onScreenCappedXStart < 0)
            onScreenCappedXStart = 0;

        // cap X to right screen border
        int onScreenCappedXEnd = onScreenUncappedXEnd;
        if (onScreenCappedXEnd > targetRenderingArea.width)
            onScreenCappedXEnd = targetRenderingArea.width;

        final byte[] targetRenderingAreaBytes = targetRenderingArea.pixels;

        final int textureWidth = textureBitmap.width;

        for (int y = onScreenCappedYStart; y < onScreenCappedYEnd; y++) {

            final int sourceBitmapScanlinePixel = ((textureBitmap.height * (y - onScreenUncappedYStart)) / onScreenUncappedHeight)
                    * textureWidth;

            int targetRenderingAreaOffset = ((y * targetRenderingArea.width) + onScreenCappedXStart) * 4;

            for (int x = onScreenCappedXStart; x < onScreenCappedXEnd; x++) {

                final int sourceBitmapPixelAddress = (sourceBitmapScanlinePixel + ((textureWidth * (x - onScreenUncappedXStart)) / onScreenUncappedWidth)) * 4;

                textureBitmap.drawPixel(sourceBitmapPixelAddress, targetRenderingAreaBytes, targetRenderingAreaOffset);

                targetRenderingAreaOffset += 4;
            }
        }
    }

    /**
     * Set the scale of the texture
     *
     * @param scale the scale of the texture
     */
    public void setScale(final double scale) {
        this.scale = scale * SCALE_MULTIPLIER;
    }

    public Point3D getLocation() {
        return coordinates[0].coordinate;
    }

}
