/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic;

import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;

import java.util.Collections;
import java.util.Set;
import java.util.WeakHashMap;

import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

public class GlowingPoint extends ForwardOrientedTexture {

    private static final int TEXTURE_RESOLUTION_PIXELS = 100;
    /**
     * A set of all existing glowing points.
     * Used to reuse textures of glowing points of the same color.
     */
    private static final Set<GlowingPoint> glowingPoints = Collections.newSetFromMap(new WeakHashMap<>());
    private final Color color;

    public GlowingPoint(final Point3D point, final double pointSize,
                        final Color color) {
        super(point, computeScale(pointSize), getTexture(color));
        this.color = color;

        synchronized (glowingPoints) {
            glowingPoints.add(this);
        }
    }


    private static double computeScale(double pointSize) {
        return pointSize / ((double) (TEXTURE_RESOLUTION_PIXELS / 50f));
    }

    /**
     * Returns a texture for a glowing point of the given color.
     * The texture is a circle with a gradient from transparent to the given color.
     */
    private static Texture getTexture(final Color color) {
        // attempt to reuse texture from existing glowing point of the same color
        synchronized (glowingPoints) {
            for (GlowingPoint glowingPoint : glowingPoints)
                if (color.equals(glowingPoint.color))
                    return glowingPoint.texture;
        }

        // existing texture not found, creating new one
        return createTexture(color);
    }

    /**
     * Creates a texture for a glowing point of the given color.
     * The texture is a circle with a gradient from transparent to the given color.
     */
    private static Texture createTexture(final Color color) {
        final Texture texture = new Texture(TEXTURE_RESOLUTION_PIXELS, TEXTURE_RESOLUTION_PIXELS, 1);
        int halfResolution = TEXTURE_RESOLUTION_PIXELS / 2;

        for (int x = 0; x < TEXTURE_RESOLUTION_PIXELS; x++)
            for (int y = 0; y < TEXTURE_RESOLUTION_PIXELS; y++) {
                int address = texture.primaryBitmap.getAddress(x, y);

                final int distanceFromCenter = (int) sqrt(pow(halfResolution - x, 2) + pow(halfResolution - y, 2));

                int alpha = 255 - ((270 * distanceFromCenter) / halfResolution);
                if (alpha < 0)
                    alpha = 0;

                texture.primaryBitmap.bytes[address] = (byte) alpha;
                address++;
                texture.primaryBitmap.bytes[address] = (byte) color.b;
                address++;
                texture.primaryBitmap.bytes[address] = (byte) color.g;
                address++;
                texture.primaryBitmap.bytes[address] = (byte) color.r;
            }

        return texture;
    }

}
