/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.texturedpolygon;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.math.Vertex;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.TextureBitmap;

import java.awt.*;

import static eu.svjatoslav.sixth.e3d.geometry.Polygon.pointWithinPolygon;
import static java.util.Arrays.sort;

/**
 * Textured polygon.
 * <p>
 *
 * <pre>
 * This is how perspective-correct texture rendering is implemented:
 * If polygon is sufficiently small, it is rendered without perspective correction.
 * Otherwise, it is sliced into smaller polygons.
 * </pre>
 */

public class TexturedPolygon extends AbstractCoordinateShape {

    public final Texture texture;

    /**
     * If <code>true</code> then polygon borders will be drawn.
     * It is used for debugging purposes.
     */
    public boolean showBorders = false;

    private double totalTextureDistance = -1;

    public TexturedPolygon(Vertex p1, Vertex p2, Vertex p3, final Texture texture) {

        super(p1, p2, p3);
        this.texture = texture;
    }

    private void computeTotalTextureDistance() {
        // compute total texture distance
        totalTextureDistance = coordinates[0].textureCoordinate.getDistanceTo(coordinates[1].textureCoordinate);
        totalTextureDistance += coordinates[0].textureCoordinate.getDistanceTo(coordinates[2].textureCoordinate);
        totalTextureDistance += coordinates[1].textureCoordinate.getDistanceTo(coordinates[2].textureCoordinate);
    }

    private void drawHorizontalLine(final PolygonBorderInterpolator line1,
                                    final PolygonBorderInterpolator line2, final int y,
                                    final RenderingContext renderBuffer,
                                    final TextureBitmap textureBitmap) {

        line1.setCurrentY(y);
        line2.setCurrentY(y);

        int x1 = line1.getX();
        int x2 = line2.getX();

        final double tx2, ty2;
        final double tx1, ty1;

        if (x1 <= x2) {

            tx1 = line1.getTX() * textureBitmap.multiplicationFactor;
            ty1 = line1.getTY() * textureBitmap.multiplicationFactor;

            tx2 = line2.getTX() * textureBitmap.multiplicationFactor;
            ty2 = line2.getTY() * textureBitmap.multiplicationFactor;

        } else {
            final int tmp = x1;
            x1 = x2;
            x2 = tmp;

            tx1 = line2.getTX() * textureBitmap.multiplicationFactor;
            ty1 = line2.getTY() * textureBitmap.multiplicationFactor;

            tx2 = line1.getTX() * textureBitmap.multiplicationFactor;
            ty2 = line1.getTY() * textureBitmap.multiplicationFactor;
        }

        final double realWidth = x2 - x1;
        final double realX1 = x1;

        if (x1 < 0)
            x1 = 0;

        if (x2 >= renderBuffer.width)
            x2 = renderBuffer.width - 1;

        int renderBufferOffset = ((y * renderBuffer.width) + x1) * 4;
        final byte[] renderBufferBytes = renderBuffer.pixels;

        final double twidth = tx2 - tx1;
        final double theight = ty2 - ty1;

        for (int x = x1; x < x2; x++) {

            final double distance = x - realX1;

            final double tx = tx1 + ((twidth * distance) / realWidth);
            final double ty = ty1 + ((theight * distance) / realWidth);

            final int textureOffset = textureBitmap.getAddress((int) tx,
                    (int) ty);

            textureBitmap.drawPixel(textureOffset, renderBufferBytes,
                    renderBufferOffset);

            renderBufferOffset += 4;
        }

    }

    @Override
    public void paint(final RenderingContext renderBuffer) {

        final Point2D projectedPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D projectedPoint2 = coordinates[1].onScreenCoordinate;
        final Point2D projectedPoint3 = coordinates[2].onScreenCoordinate;

        projectedPoint1.roundToInteger();
        projectedPoint2.roundToInteger();
        projectedPoint3.roundToInteger();

        if (mouseInteractionController != null)
            if (renderBuffer.getMouseEvent() != null)
                if (pointWithinPolygon(
                        renderBuffer.getMouseEvent().coordinate, projectedPoint1,
                        projectedPoint2, projectedPoint3))
                    renderBuffer.setCurrentObjectUnderMouseCursor(mouseInteractionController);

        // Show polygon boundaries (for debugging)
        if (showBorders)
            showBorders(renderBuffer);

        // find top-most point
        int yTop = (int) projectedPoint1.y;

        if (projectedPoint2.y < yTop)
            yTop = (int) projectedPoint2.y;

        if (projectedPoint3.y < yTop)
            yTop = (int) projectedPoint3.y;

        if (yTop < 0)
            yTop = 0;

        // find bottom-most point
        int yBottom = (int) projectedPoint1.y;

        if (projectedPoint2.y > yBottom)
            yBottom = (int) projectedPoint2.y;

        if (projectedPoint3.y > yBottom)
            yBottom = (int) projectedPoint3.y;

        if (yBottom >= renderBuffer.height)
            yBottom = renderBuffer.height - 1;

        // paint
        double totalVisibleDistance = projectedPoint1.getDistanceTo(projectedPoint2);
        totalVisibleDistance += projectedPoint1.getDistanceTo(projectedPoint3);
        totalVisibleDistance += projectedPoint2.getDistanceTo(projectedPoint3);

        if (totalTextureDistance == -1)
            computeTotalTextureDistance();
        final double scaleFactor = (totalVisibleDistance / totalTextureDistance) * 1.2d;

        final TextureBitmap zoomedBitmap = texture.getZoomedBitmap(scaleFactor);

        final PolygonBorderInterpolator[] is = new PolygonBorderInterpolator[]{
                new PolygonBorderInterpolator(), new PolygonBorderInterpolator(),
                new PolygonBorderInterpolator()};

        is[0].setPoints(projectedPoint1, projectedPoint2,
                coordinates[0].textureCoordinate,
                coordinates[1].textureCoordinate);
        is[1].setPoints(projectedPoint1, projectedPoint3,
                coordinates[0].textureCoordinate,
                coordinates[2].textureCoordinate);
        is[2].setPoints(projectedPoint2, projectedPoint3,
                coordinates[1].textureCoordinate,
                coordinates[2].textureCoordinate);

        sort(is);

        for (int y = yTop; y < yBottom; y++)
            if (is[0].containsY(y)) {

                if (is[1].containsY(y))
                    drawHorizontalLine(is[0], is[1], y, renderBuffer,
                            zoomedBitmap);
                else if (is[2].containsY(y))
                    drawHorizontalLine(is[0], is[2], y, renderBuffer,
                            zoomedBitmap);
            } else if (is[1].containsY(y))
                if (is[2].containsY(y))
                    drawHorizontalLine(is[1], is[2], y, renderBuffer,
                            zoomedBitmap);

    }

    private void showBorders(final RenderingContext renderBuffer) {

        final Point2D projectedPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D projectedPoint2 = coordinates[1].onScreenCoordinate;
        final Point2D projectedPoint3 = coordinates[2].onScreenCoordinate;

        renderBuffer.graphics.setColor(Color.YELLOW);
        renderBuffer.graphics.drawLine((int) projectedPoint1.x,
                (int) projectedPoint1.y, (int) projectedPoint2.x,
                (int) projectedPoint2.y);
        renderBuffer.graphics.drawLine((int) projectedPoint3.x,
                (int) projectedPoint3.y, (int) projectedPoint2.x,
                (int) projectedPoint2.y);
        renderBuffer.graphics.drawLine((int) projectedPoint1.x,
                (int) projectedPoint1.y, (int) projectedPoint3.x,
                (int) projectedPoint3.y);
    }

}
