/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.base;

import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.gui.UserRelativityTracker;
import eu.svjatoslav.sixth.e3d.gui.humaninput.MouseInteractionController;
import eu.svjatoslav.sixth.e3d.math.Transform;
import eu.svjatoslav.sixth.e3d.math.TransformsStack;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.RenderAggregator;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractShape;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.line.Line;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.solidpolygon.SolidPolygon;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.texturedpolygon.TexturedPolygon;
import eu.svjatoslav.sixth.e3d.renderer.raster.slicer.Slicer;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

public class AbstractCompositeShape extends AbstractShape {
    private final List<SubShape> originalSubShapes = new ArrayList<>();
    private final UserRelativityTracker relativityTracker;
    double currentSliceFactor = 5;
    private List<AbstractShape> renderedSubShapes = new ArrayList<>();
    private boolean slicingOutdated = true;
    private Transform transform;

    public AbstractCompositeShape() {
        this(new Transform());
    }

    public AbstractCompositeShape(final Point3D location) {
        this(new Transform(location));
    }

    public AbstractCompositeShape(final Transform transform) {
        this.transform = transform;
        relativityTracker = new UserRelativityTracker();
    }

    public void addShape(final AbstractShape shape) {
        addShape(shape, null);
    }

    public void addShape(final AbstractShape shape, final String groupId) {
        final SubShape subShape = new SubShape(shape);
        subShape.setGroup(groupId);
        subShape.setVisible(true);
        originalSubShapes.add(subShape);
        slicingOutdated = true;
    }

    /**
     * This method should be overridden by anyone wanting to customize shape
     * before it is rendered.
     */
    public void beforeTransformHook(final TransformsStack transformPipe,
                                    final RenderingContext context) {
    }

    public Point3D getLocation() {
        return transform.getTranslation();
    }

    public List<SubShape> getOriginalSubShapes() {
        return originalSubShapes;
    }

    public UserRelativityTracker getRelativityTracker() {
        return relativityTracker;
    }

    public void hideGroup(final String groupIdentifier) {
        originalSubShapes.stream().filter(subShape -> subShape.matchesGroup(groupIdentifier)).forEach(subShape -> {
            subShape.setVisible(false);
            slicingOutdated = true;
        });
    }

    private boolean isReslicingNeeded(double proposedNewSliceFactor, double currentSliceFactor) {

        if (slicingOutdated)
            return true;

        // reslice if there is significant difference between proposed and current slice factor
        if (proposedNewSliceFactor > currentSliceFactor) {
            final double tmp = proposedNewSliceFactor;
            proposedNewSliceFactor = currentSliceFactor;
            currentSliceFactor = tmp;
        }

        return (currentSliceFactor / proposedNewSliceFactor) > 1.5d;
    }

    public void removeGroup(final String groupIdentifier) {
        final java.util.Iterator<SubShape> iterator = originalSubShapes
                .iterator();

        while (iterator.hasNext()) {
            final SubShape subShape = iterator.next();
            if (subShape.matchesGroup(groupIdentifier)) {
                iterator.remove();
                slicingOutdated = true;
            }
        }
    }

    public List<SubShape> getGroup(final String groupIdentifier) {
        return originalSubShapes.stream().filter(
                        subShape -> subShape.matchesGroup(groupIdentifier))
                .collect(Collectors.toList());
    }

    private void resliceIfNeeded() {

        final double proposedSliceFactor = relativityTracker.proposeSliceFactor();

        if (isReslicingNeeded(proposedSliceFactor, currentSliceFactor)) {
            currentSliceFactor = proposedSliceFactor;
            reslice();
        }
    }

    /**
     * Paint solid elements of this composite shape into given color.
     */
    public void setColor(final Color color) {
        for (final SubShape subShape : getOriginalSubShapes()) {
            final AbstractShape shape = subShape.getShape();

            if (shape instanceof SolidPolygon)
                ((SolidPolygon) shape).setColor(color);

            if (shape instanceof Line)
                ((Line) shape).color = color;
        }
    }

    public void setGroupForUngrouped(final String groupIdentifier) {
        originalSubShapes.stream().filter(SubShape::isUngrouped).forEach(subShape -> subShape.setGroup(groupIdentifier));
    }

    @Override
    public void setMouseInteractionController(
            final MouseInteractionController mouseInteractionController) {
        super.setMouseInteractionController(mouseInteractionController);

        for (final SubShape subShape : originalSubShapes)
            subShape.getShape().setMouseInteractionController(
                    mouseInteractionController);

        slicingOutdated = true;

    }

    public void setTransform(final Transform transform) {
        this.transform = transform;
    }

    public void showGroup(final String groupIdentifier) {
        originalSubShapes.stream().filter(subShape -> subShape.matchesGroup(groupIdentifier)).forEach(subShape -> {
            subShape.setVisible(true);
            slicingOutdated = true;
        });
    }

    private void reslice() {
        slicingOutdated = false;

        final List<AbstractShape> result = new ArrayList<>();

        final Slicer slicer = new Slicer(currentSliceFactor);
        originalSubShapes.stream().filter(SubShape::isVisible).forEach(subShape -> {
            if (subShape.getShape() instanceof TexturedPolygon)
                slicer.slice((TexturedPolygon) subShape.getShape());
            else
                result.add(subShape.getShape());
        });

        result.addAll(slicer.getResult());

        renderedSubShapes = result;
    }

    @Override
    public void transform(final TransformsStack transformPipe,
                          final RenderAggregator aggregator, final RenderingContext context) {

        // add current composite shape transform to the end of the transform
        // pipeline
        transformPipe.addTransform(transform);

        relativityTracker.analyze(transformPipe, context);

        beforeTransformHook(transformPipe, context);

        // hack, to get somewhat perspective correct textures
        resliceIfNeeded();

        // transform rendered subshapes
        for (final AbstractShape shape : renderedSubShapes)
            shape.transform(transformPipe, aggregator, context);

        transformPipe.dropTransform();
    }

}
