/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.textcanvas;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.solidpolygon.SolidPolygon;

import java.awt.*;

import static eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.solidpolygon.SolidPolygon.drawPolygon;
import static eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.textcanvas.TextCanvas.FONT_CHAR_HEIGHT;
import static eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.textcanvas.TextCanvas.FONT_CHAR_WIDTH;
import static java.lang.String.valueOf;

/**
 * Represents a single character on the text canvas.
 */
public class CanvasCharacter extends AbstractCoordinateShape {

    private static final int MAX_FONT_SIZE = 500;

    /**
     * Cached fonts.
     */
    private static final Font[] fonts = new Font[MAX_FONT_SIZE];

    /**
     * The character to be rendered.
     */
    private char value;

    /**
     * The foreground color of the character.
     */
    private eu.svjatoslav.sixth.e3d.renderer.raster.Color foregroundColor;

    /**
     * The background color of the character.
     */
    private eu.svjatoslav.sixth.e3d.renderer.raster.Color backgroundColor;

    public CanvasCharacter(final Point3D centerLocation, final char character,
                           final eu.svjatoslav.sixth.e3d.renderer.raster.Color foregroundColor,
                           final eu.svjatoslav.sixth.e3d.renderer.raster.Color backgroundColor) {

        // There are 5 coordinates: center, upper left, upper right, lower right, lower left
        super(5);

        value = character;
        this.foregroundColor = foregroundColor;
        this.backgroundColor = backgroundColor;


        coordinates[0].coordinate = centerLocation;

        final double halfWidth = FONT_CHAR_WIDTH / 2d;
        final double halfHeight = FONT_CHAR_HEIGHT / 2d;

        // upper left
        coordinates[1].coordinate = centerLocation.clone().translateX(-halfWidth)
                .translateY(-halfHeight);

        // upper right
        coordinates[2].coordinate = centerLocation.clone().translateX(halfWidth)
                .translateY(-halfHeight);

        // lower right
        coordinates[3].coordinate = centerLocation.clone().translateX(halfWidth)
                .translateY(halfHeight);

        // lower left
        coordinates[4].coordinate = centerLocation.clone().translateX(-halfWidth)
                .translateY(halfHeight);
    }

    /**
     * Returns a font of the specified size.
     * <p>
     * If the font of the specified size is already cached, it will be
     * returned. Otherwise, a new font will be created, cached and returned.
     *
     * @param size the size of the font
     * @return the font
     */
    public static Font getFont(final int size) {
        if (fonts[size] != null)
            return fonts[size];

        final Font font = new Font("Courier", Font.BOLD, size);
        fonts[size] = font;
        return font;
    }

    /**
     * Returns color of the background.
     */
    public eu.svjatoslav.sixth.e3d.renderer.raster.Color getBackgroundColor() {
        return backgroundColor;
    }

    /**
     * Sets color of the background.
     */
    public void setBackgroundColor(
            final eu.svjatoslav.sixth.e3d.renderer.raster.Color backgroundColor) {
        this.backgroundColor = backgroundColor;
    }

    /**
     * Returns color of the foreground.
     *
     * @return the color
     */
    public eu.svjatoslav.sixth.e3d.renderer.raster.Color getForegroundColor() {
        return foregroundColor;
    }

    /**
     * Sets color of the foreground.
     *
     * @param foregroundColor the color
     */
    public void setForegroundColor(
            final eu.svjatoslav.sixth.e3d.renderer.raster.Color foregroundColor) {
        this.foregroundColor = foregroundColor;
    }

    /**
     * Paints the character on the screen.
     * @param renderingContext the rendering context
     */
    @Override
    public void paint(final RenderingContext renderingContext) {

        // Draw background rectangle first. It is composed of two triangles.
        drawPolygon(renderingContext,
                coordinates[1].onScreenCoordinate,
                coordinates[2].onScreenCoordinate,
                coordinates[3].onScreenCoordinate,
                mouseInteractionController,
                backgroundColor);

        drawPolygon(renderingContext,
                coordinates[1].onScreenCoordinate,
                coordinates[3].onScreenCoordinate,
                coordinates[4].onScreenCoordinate,
                mouseInteractionController,
                backgroundColor);

        final int desiredFontSize = (int) ((renderingContext.width * 4.5) / onScreenZ);

        // do not render too large characters
        if (desiredFontSize >= MAX_FONT_SIZE)
            return;

        final Point2D onScreenLocation = coordinates[0].onScreenCoordinate;

        // screen borders check
        if (onScreenLocation.x < 0)
            return;
        if (onScreenLocation.y < 0)
            return;

        if (onScreenLocation.x > renderingContext.width)
            return;
        if (onScreenLocation.y > renderingContext.height)
            return;

        // draw the character
        renderingContext.graphics.setFont(getFont(desiredFontSize));
        renderingContext.graphics.setColor(foregroundColor.toAwtColor());
        renderingContext.graphics.drawString(
                valueOf(value),
                (int) onScreenLocation.x - (int) (desiredFontSize / 3.2),
                (int) onScreenLocation.y + (int) (desiredFontSize / 2.5));

    }

    public void setValue(final char value) {
        this.value = value;
    }

}
